-----------------------------------------------------------------------------------------------
-- Client Lua Script for HousingRepairWatcher
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "HousingLib"
require "GameLib"
 
-----------------------------------------------------------------------------------------------
-- HousingRepairWatcher Module Definition
-----------------------------------------------------------------------------------------------
local HousingRepairWatcher = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
HousingRepairWatcher.CurrentVersion = "1.0.4"

-- Control prefixes
-- Plot button prefix
HousingRepairWatcher.PlotControlPrefix = "btnPlot."
-- Plot Name prefix
HousingRepairWatcher.PlotNameControlPrefix = "txtPlotName."
-- Time Remaining prefix
HousingRepairWatcher.PlotTimeRemainingControlPrefix = "lblTimeRemaining."
-- Occupied Sprite prefix
HousingRepairWatcher.PlotOccupiedSpritePrefix = "spOccupied."
-- Disabled Sprite prefix
HousingRepairWatcher.PlotDisabledSpritePrefix = "spDisabled."
-- Disabled Sprite prefix
HousingRepairWatcher.PlotRepairBarPrefix = "pbRepair."


-- Helper variables
HousingRepairWatcher.CurrentCharacterKey = ""
HousingRepairWatcher.FullyLoaded = false
HousingRepairWatcher.ConfigLoaded = false
HousingRepairWatcher.AlertLoaded = false

HousingRepairWatcher.CurrentAlertCount = 0
HousingRepairWatcher.bDismissAlertClick = true

-----------------------------------------------------------------------------------------------
-- Save Data
-----------------------------------------------------------------------------------------------
HousingRepairWatcher.AccountCharacters = {}
HousingRepairWatcher.Settings = {}
HousingRepairWatcher.Configuration = {}

-- Print Debug Messages
HousingRepairWatcher.Configuration["bDisplayDebugMessages"] = {}
HousingRepairWatcher.Configuration["bDisplayDebugMessages"].ConfigDisplay = "Show Debug"
HousingRepairWatcher.Configuration["bDisplayDebugMessages"].ConfigValue = false
-- Auto Display
HousingRepairWatcher.Configuration["bAutoDisplay"] = {}
HousingRepairWatcher.Configuration["bAutoDisplay"].ConfigDisplay = "Auto Display (at home)"
HousingRepairWatcher.Configuration["bAutoDisplay"].ConfigValue = true
-- Auto Repair
HousingRepairWatcher.Configuration["bAutoRepair"] = {}
HousingRepairWatcher.Configuration["bAutoRepair"].ConfigDisplay = "Auto Repair"
HousingRepairWatcher.Configuration["bAutoRepair"].ConfigValue = false

-- Enable Alerts
HousingRepairWatcher.Configuration["bAlertEnabled"] = {}
HousingRepairWatcher.Configuration["bAlertEnabled"].ConfigDisplay = "Enable Alerts"
HousingRepairWatcher.Configuration["bAlertEnabled"].ConfigValue = true

-- Alert Times
HousingRepairWatcher.Configuration["b60MinuteAlert"] = {}
HousingRepairWatcher.Configuration["b60MinuteAlert"].ConfigDisplay = "60 Min Alerts"
HousingRepairWatcher.Configuration["b60MinuteAlert"].ConfigValue = true
HousingRepairWatcher.Configuration["b45MinuteAlert"] = {}
HousingRepairWatcher.Configuration["b45MinuteAlert"].ConfigDisplay = "45 Min Alerts"
HousingRepairWatcher.Configuration["b45MinuteAlert"].ConfigValue = true
HousingRepairWatcher.Configuration["b30MinuteAlert"] = {}
HousingRepairWatcher.Configuration["b30MinuteAlert"].ConfigDisplay = "30 Min Alerts"
HousingRepairWatcher.Configuration["b30MinuteAlert"].ConfigValue = true
HousingRepairWatcher.Configuration["b15MinuteAlert"] = {}
HousingRepairWatcher.Configuration["b15MinuteAlert"].ConfigDisplay = "15 Min Alerts"
HousingRepairWatcher.Configuration["b15MinuteAlert"].ConfigValue = true
HousingRepairWatcher.Configuration["b5MinuteAlert"] = {}
HousingRepairWatcher.Configuration["b5MinuteAlert"].ConfigDisplay = "5 Min Alerts"
HousingRepairWatcher.Configuration["b5MinuteAlert"].ConfigValue = true

HousingRepairWatcher.AlertBlacklist = {}

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function HousingRepairWatcher:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function HousingRepairWatcher:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "Housing Repair Watcher"
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

-----------------------------------------------------------------------------------------------
-- Saving / Loading
----------------------------------------------------------------------------------------------- 
function HousingRepairWatcher:OnSave(eLevel)
	if eLevel == GameLib.CodeEnumAddonSaveLevel.Account then
		-- Save account characters including alts
		return HousingRepairWatcher.AccountCharacters
	elseif eLevel == GameLib.CodeEnumAddonSaveLevel.Character then
		-- Save addon settings
		return HousingRepairWatcher.Configuration
	end
end

function HousingRepairWatcher:OnRestore(eLevel, tData)
	if tData ~= nil then
		if eLevel == GameLib.CodeEnumAddonSaveLevel.Account then
			-- Load Account Character data
			--HousingRepairWatcher.AccountCharacters = tData
			HousingRepairWatcher.AccountCharacters = {}

			-- For each character save data
		    for SaveKey, tSaveData in pairs(tData) do
		    	-- Construct a new save data to store
		    	local tPlayerSaveData = {}
		    	
		    	tPlayerSaveData.nPlayerId = tSaveData.nPlayerId
		    	tPlayerSaveData.strPlayerName = tSaveData.strPlayerName
		    	tPlayerSaveData.strPlayerRealm = tSaveData.strPlayerRealm

		    	tPlayerSaveData.tPlots = {}

		    	-- For each plot in the character's save, add it to
		    	-- the character
		    	for idx, tPlot in pairs(tSaveData.tPlots) do
		    		local tPlotSaveData = {}

		    		tPlotSaveData.strPlotName = tPlot.strPlotName
					tPlotSaveData.nPlotIndex = tPlot.nPlotIndex

					-- Timestamp info
					tPlotSaveData.timestamp = {}
						tPlotSaveData.timestamp.nHour = tPlot.timestamp.nHour
						tPlotSaveData.timestamp.strFormattedTime = tPlot.timestamp.strFormattedTime
						tPlotSaveData.timestamp.nYear = tPlot.timestamp.nYear
						tPlotSaveData.timestamp.nSecond = tPlot.timestamp.nSecond
						tPlotSaveData.timestamp.nDayOfWeek = tPlot.timestamp.nDayOfWeek
						tPlotSaveData.timestamp.nDay = tPlot.timestamp.nDay
						tPlotSaveData.timestamp.nMinute = tPlot.timestamp.nMinute
						tPlotSaveData.timestamp.nMonth = tPlot.timestamp.nMonth

					tPlotSaveData.fPlotRemainingHours = tPlot.fPlotRemainingHours
					tPlotSaveData.eUpkeepType = tPlot.eUpkeepType
					tPlotSaveData.bNeedsRepair = tPlot.bNeedsRepair	
					tPlotSaveData.nPlotChargesRemaining = tPlot.nPlotChargesRemaining

					tPlayerSaveData.tPlots[idx] = tPlotSaveData
		    	end
		    	
		    	-- Add the save to the addon's character list
		    	HousingRepairWatcher.AccountCharacters[SaveKey] = tPlayerSaveData
		    end
		elseif eLevel == GameLib.CodeEnumAddonSaveLevel.Character then
			
			-- Load addon settings
			if HousingRepairWatcher.Configuration == nil then
				HousingRepairWatcher.Configuration = {}
			end

			for key, tSetting in pairs(tData) do
				if HousingRepairWatcher.Configuration[key] ~= nil then
					tConfigItem = {}
					tConfigItem.ConfigDisplay = tSetting.ConfigDisplay
					tConfigItem.ConfigValue = tSetting.ConfigValue
			    	HousingRepairWatcher.Configuration[key] = tConfigItem
		    	end
		    end
		end
	end
end
-----------------------------------------------------------------------------------------------
-- HousingRepairWatcher OnLoad
-----------------------------------------------------------------------------------------------
function HousingRepairWatcher:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("HousingRepairWatcher.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- HousingRepairWatcher OnDocLoaded
-----------------------------------------------------------------------------------------------
function HousingRepairWatcher:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		--self:PrintDebugMessage("OnDocLoaded")
	    HousingRepairWatcher.wndMain = Apollo.LoadForm(self.xmlDoc, "frmRepairWatcherForm", nil, self)
		if HousingRepairWatcher.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    HousingRepairWatcher.wndMain:Show(false, true)

		
	    -- Hide the player list
	    local wPlayerList = HousingRepairWatcher.wndMain:FindChild("wPlayerListFrame")
	    if wPlayerList ~= nil then
	    	wPlayerList:Show(false)
	    end

		Apollo.RegisterSlashCommand("hrw", "OnHousingRepairWatcherOn", self)
		Apollo.RegisterEventHandler("ZoneCompletionUpdated", "OnZoneCompletionUpdated",  self)
		Apollo.RegisterTimerHandler("HousingRepairWatcherTimer", "OnUpdateTimer", self)
		Apollo.RegisterEventHandler("ShowWatcherWindow", "ShowWatcherWindow", self)
		Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)


		-- Start the timer for the addon
		-- Timer object
		self.UpdateTimer = Apollo.CreateTimer("HousingRepairWatcherTimer", 1.0, true)
		Apollo.StartTimer("HousingRepairWatcherTimer")

		self.FullyLoaded = true;

		-- Load configuration form file
		self.xmlConfig = XmlDoc.CreateFromFile("HousingRepairWatcherConfig.xml")
		self.xmlConfig:RegisterCallback("OnConfigLoaded", self)

		-- Load configuration form file
		self.xmlAlert = XmlDoc.CreateFromFile("HousingRepairWatcherAlert.xml")
		self.xmlAlert:RegisterCallback("OnAlertLoaded", self)
	end
end

function HousingRepairWatcher:OnConfigLoaded()
	self:PrintDebugMessage("In OnConfigLoaded")
	if self.xmlConfig ~= nil and self.xmlConfig:IsLoaded() then
		HousingRepairWatcher.wndConfig = Apollo.LoadForm(self.xmlConfig, "frmHousingRepairWatcherConfig", nil, self)
		if HousingRepairWatcher.wndConfig == nil then
			Apollo.AddAddonErrorText(self, "Could not load the configuration window for some reason.")
			Print("Could not load the configuration window for some reason.")
			return
		end
	end

	HousingRepairWatcher.wndConfig:Show(false, true)


	self.ConfigLoaded = true;
end

function HousingRepairWatcher:OnAlertLoaded()
	self:PrintDebugMessage("In OnAlertLoaded")
	if self.xmlAlert ~= nil and self.xmlAlert:IsLoaded() then
		self.wndAlert = Apollo.LoadForm(self.xmlAlert, "frmHousingRepairWatcherAlertContainer", nil, self)
		if self.wndAlert == nil then
			Apollo.AddAddonErrorText(self, "Could not load the configuration window for some reason.")
			return
		end
	end

	self.wndAlert:Show(false, true)

	self.AlertLoaded = true;
	
end

function HousingRepairWatcher:OnInterfaceMenuListHasLoaded()
	--{"ShowOptions", "", ""}
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "HousingRepairWatcher", {"ShowWatcherWindow", "", ""})
end

-----------------------------------------------------------------------------------------------
-- HousingRepairWatcher Events
-----------------------------------------------------------------------------------------------

function HousingRepairWatcher:HousingRepairWatcherConfigItemChecked( wndHandler, wndControl, eMouseButton )
	if wndControl ~= nil then
		HousingRepairWatcher.Configuration[wndControl:GetData()].ConfigValue = true
	end
end

function HousingRepairWatcher:HousingRepairWatcherConfigItemUnChecked( wndHandler, wndControl, eMouseButton )
	if wndControl ~= nil then
		HousingRepairWatcher.Configuration[wndControl:GetData()].ConfigValue = false
	end
end

function HousingRepairWatcher:HousingRepairWatcherConfigItemTextChanged( wndHandler, wndControl, strText )
	if wndControl ~= nil then
		HousingRepairWatcher.Configuration[wndControl:GetData()].ConfigValue = strText
	end
end

function HousingRepairWatcher:btnSaveConfigurationClicked( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY )
	-- Close the config window
	HousingRepairWatcher.wndConfig:Show(false)
end

function HousingRepairWatcher:frmAlertItemMouseEnter( wndHandler, wndControl, x, y )
	if wndControl ~= nil then
		-- Get a handle on the X overlay and ensure it's visible
		local lblDismissAlert = wndControl:FindChild("lblClose")

		HousingRepairWatcher.bDismissAlertClick = true

		if lblDismissAlert ~= nil then
			lblDismissAlert:Show(true)
		end
	end
end

function HousingRepairWatcher:frmAlertItemMouseExit( wndHandler, wndControl, x, y )
	if wndControl ~= nil then
		-- Get a handle on the X overlay and ensure it's visible
		local lblDismissAlert = wndControl:FindChild("lblClose")
		if lblDismissAlert ~= nil then
			lblDismissAlert:Show(false)
		end
	end
end

function HousingRepairWatcher:frmAlertItemMouseUp( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY )
	-- dismiss alert
	if HousingRepairWatcher.bDismissAlertClick == false then
		return
	end

	wndControl:Destroy()

	if next(self.wndAlert:GetChildren()) == nil then
		self.wndAlert:Show(false)
	else
		self.wndAlert:ArrangeChildrenVert()
	end


end

function HousingRepairWatcher:btnTestAlertsClicked( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY )
	self:TestAlerts()
end

function HousingRepairWatcher:TestAlerts()
	for nPlayerIndex, tPlayerSaveData in pairs (HousingRepairWatcher.AccountCharacters) do
		for nPlotIndex, tPlotSaveData in pairs(tPlayerSaveData.tPlots) do
			local nPlotHoursRemaining = self:CalculateHoursRemaining(tPlotSaveData.timestamp, tPlotSaveData.fPlotRemainingHours)
			local nPlotMinutesRemaining = math.floor(nPlotHoursRemaining * 60)

			-- Loop through each alert time and check if the current plot is in the blacklist. A blacklisted plot
			-- is one that has already had an alert for the alert time. This is used to ensure plots don't constantly
			-- pop up alerts each second. We'll retrieve a list of blacklist times and use that to determine if we
			-- need to show the alert or not.

			-- We key our blacklist with "<player realm><player name><plot index>.<minutes>" to ensure unique blacklist lookups
			local blacklistKeyPrefix = tPlayerSaveData.strPlayerRealm .. tPlayerSaveData.strPlayerName .. tostring(nPlotIndex) .. "."
			local wAlertItem = self:CreateExpirationAlert(tPlayerSaveData, tPlotSaveData, blacklistKeyPrefix .. tostring(nPlotMinutesRemaining))
			HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] = true

			HousingRepairWatcher.CurrentAlertCount = HousingRepairWatcher.CurrentAlertCount + 1
		end
	end
end

function HousingRepairWatcher:OnConfigure()
	
	-- Open the configuration window if it exists and is loaded
	if HousingRepairWatcher.wndConfig ~= nil and HousingRepairWatcher.wndConfig:IsVisible() == false then
		HousingRepairWatcher.wndConfig:Show(true, true)
		-- Load all of the settings into the configuration window
	    local wSettingContainer = HousingRepairWatcher.wndConfig:FindChild("wConfigurationItemContainer")

		wSettingContainer:DestroyChildren()

		-- Loop through each setting and add the appropriate setting control to the setting container
		table.sort(HousingRepairWatcher.Configuration)
		for key, setting in pairs(HousingRepairWatcher.Configuration) do
		--for key, setting in pairs(HousingRepairWatcher.Configuration) do
			if type(setting.ConfigValue) == 'boolean' then
				local wSettingItem = Apollo.LoadForm(self.xmlConfig, "frmBoolConfigItem", wSettingContainer, self)
				local wSettingControl = wSettingItem:FindChild("ConfigItem")
				wSettingControl:SetData(key)
				wSettingControl:SetText(setting.ConfigDisplay)
				wSettingControl:SetCheck(setting.ConfigValue)
			elseif type(setting.ConfigValue) == "string" then
				local wSettingItem = Apollo.LoadForm(self.xmlConfig, "frmStringConfigItem", wSettingContainer, self)
				local wSettingControl = wSettingItem:FindChild("ConfigItem")
				local wSettingLabel = wSettingItem:FindChild("lblSettingName")
				wSettingControl:SetData(key)
				wSettingLabel:SetText(setting.ConfigDisplay)
				wSettingControl:SetText(setting.ConfigValue)
			end
		end

		wSettingContainer:ArrangeChildrenVert()
	end
end

-- on SlashCommand "/hrw"
function HousingRepairWatcher:OnHousingRepairWatcherOn(strCmd, strParam)
	
	if strParam ~= nil and strParam ~= "" then
		if string.lower(strParam) == "config" then
			self:OnConfigure()
		elseif string.lower(strParam) == "test" then
			self:TestAlerts()
		end
	else
		self:ShowWatcherWindow()
	end
end

function HousingRepairWatcher:ShowWatcherWindow()
if self.FullyLoaded == false then
		return
	end

	-- get offsets. We'll use these to position the window to the last position	
	
	--HousingRepairWatcher.wndMain:SetAnchorOffsets(HousingRepairWatcher.Settings["nWindowLeft"], HousingRepairWatcher.Settings["nWindowTop"], HousingRepairWatcher.Settings["nWindowRight"], HousingRepairWatcher.Settings["nWindowRight"])

	local bInHousing = HousingLib:IsHousingWorld()
	local bInPlayerResidence = HousingLib:IsOnMyResidence()

	-- Show the window and populate the data
	if HousingRepairWatcher.wndMain:IsVisible() == false then
		HousingRepairWatcher.wndMain:Invoke()
	end



	-- If there isn't a current character key, we set it to the current player.
	if HousingRepairWatcher.CurrentCharacterKey == nil or HousingRepairWatcher.CurrentCharacterKey == "" then
		-- create the character key for the current player
		HousingRepairWatcher.CurrentCharacterKey = GameLib.GetRealmName() .. "-" .. GameLib.GetPlayerUnit():GetTitleOrName()

		-- if the account list doesn't have an entry for the character key (current player), create it
		if HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey] == nil then
			HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey] = self:CreateOrUpdateCharacterSaveData()
		end
	end
		

	-- Update player name to player name from account list
	local playerTitleControl = HousingRepairWatcher.wndMain:FindChild("PlayerTitle")
	playerTitleControl:SetText(HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey].strPlayerName)
	
	-- Ensure that the plots table is created. This should only ever actually be an issue the very first
	-- time a player logs in after installing the addon
	if HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey].tPlots == nil then
		HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey].tPlots = {}
	end

	--self:PrintTable(HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey])

	-- Loop through the player's plots. We don't care
	-- about the house plot so start at index 2 since index 1
	-- is always the player's house

	-- Check if current key is nil. Depending on zone transition and load times, this can
	-- happen if it's unable to get the player's information
	if HousingRepairWatcher.CurrentCharacterKey == nil then
		return
	end

	for i=2,7 do
		local plot = nil
	
		-- If it's the current user, check if they are in their housing. We want to get
		-- the lastest data from the API if it's available. Otherwise use saved data
		if HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey].strPlayerName == GameLib.GetPlayerUnit():GetTitleOrName() then
			-- if the user is in their housing zone, grab the latest info for the
			-- plot
			if bInHousing == true and bInPlayerResidence == true then
				plot = self:CreateOrUpdatePlotSaveData(HousingLib.GetPlot(i))
				-- Save the latest plot info to the account table
				HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey].tPlots[i] = plot
			else
				plot = HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey].tPlots[i]
			end
		else
			plot = HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey].tPlots[i]
		end

		-- Update the plot Display
		self:UpdatePlot(i, plot)
		
	end
end

-- When changing zones
function HousingRepairWatcher:OnZoneCompletionUpdated()
	-- Auto-load the repair watcher when entering the housing zone
	if HousingLib:IsHousingWorld() == true and HousingLib:IsOnMyResidence() and HousingRepairWatcher.Configuration["bDisplayDebugMessages"].ConfigValue == true then
		self:OnHousingRepairWatcherOn(self)
	end
end

function HousingRepairWatcher:OnUpdateTimer()
	if self.FullyLoaded == false then
		return
	end

	if HousingRepairWatcher.wndMain:IsVisible() == true then
		self:OnHousingRepairWatcherOn()
	end

	-- Determine if anything needs repairs if the user has auto-repair active
	if HousingRepairWatcher.Configuration["bAutoRepair"] ~= nil and HousingRepairWatcher.Configuration["bAutoRepair"].ConfigValue == true then
		-- Check if the user is the current user
		if HousingRepairWatcher.CurrentCharacterKey ~= nil or HousingRepairWatcher.CurrentCharacterKey ~= "" then
			local bInHousing = HousingLib:IsHousingWorld()
			local bInPlayerResidence = HousingLib:IsOnMyResidence()
			local playerSaveData = HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey]
			
			if HousingRepairWatcher.AccountCharacters[HousingRepairWatcher.CurrentCharacterKey] ~= nil and bInHousing == true and bInPlayerResidence == true then
				if playerSaveData.tPlots ~= nil then
					for idx, tPlot in pairs(playerSaveData.tPlots) do
						self:PrintDebugMessage("Repairing: " .. tPlot.strPlotName)
						HousingLib.RepairPlot(idx)
					end
				end
			end
		end
	end

	-- If alerts are on, display a message on any plot that is about to expire
	if HousingRepairWatcher.Configuration["bAlertEnabled"] ~= nil and HousingRepairWatcher.Configuration["bAlertEnabled"].ConfigValue == true then
		-- We want to loop through each account and check their plots for expiration
		for nPlayerIndex, tPlayerSaveData in pairs (HousingRepairWatcher.AccountCharacters) do
			for nPlotIndex, tPlotSaveData in pairs(tPlayerSaveData.tPlots) do
				local nPlotHoursRemaining = self:CalculateHoursRemaining(tPlotSaveData.timestamp, tPlotSaveData.fPlotRemainingHours)
				local nPlotMinutesRemaining = math.floor(nPlotHoursRemaining * 60)

				-- Loop through each alert time and check if the current plot is in the blacklist. A blacklisted plot
				-- is one that has already had an alert for the alert time. This is used to ensure plots don't constantly
				-- pop up alerts each second. We'll retrieve a list of blacklist times and use that to determine if we
				-- need to show the alert or not.

				-- We key our blacklist with "<player realm><player name><plot index>.<minutes>" to ensure unique blacklist lookups
				local blacklistKeyPrefix = tPlayerSaveData.strPlayerRealm .. tPlayerSaveData.strPlayerName .. tostring(nPlotIndex) .. "."

				-- Only show an alert if the upkeep is timed
				if tPlotSaveData.eUpkeepType == HousingLib.HousingUpkeepType.Timed then
					if nPlotMinutesRemaining  == 60 and HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] == nil then
						-- create 60 minute alert
						local wAlertItem = self:CreateExpirationAlert(tPlayerSaveData, tPlotSaveData, blacklistKeyPrefix .. tostring(nPlotMinutesRemaining))
						HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] = true
					elseif nPlotMinutesRemaining == 45 and HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] == nil then
						local wAlertItem = self:CreateAlert(tPlayerSaveData, tPlotSaveData, blacklistKeyPrefix .. tostring(nPlotMinutesRemaining))
						HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] = true
					elseif nPlotMinutesRemaining == 30 and HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] == nil then
						local wAlertItem = self:CreateAlert(tPlayerSaveData, tPlotSaveData, blacklistKeyPrefix .. tostring(nPlotMinutesRemaining))
						HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] = true
					elseif nPlotMinutesRemaining == 15 and HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] == nil then
						local wAlertItem = self:CreateAlert(tPlayerSaveData, tPlotSaveData, blacklistKeyPrefix .. tostring(nPlotMinutesRemaining))
						HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] = true
					elseif nPlotMinutesRemaining == 5 and HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] == nil then
						local wAlertItem = self:CreateAlert(tPlayerSaveData, tPlotSaveData, blacklistKeyPrefix .. tostring(nPlotMinutesRemaining))
						HousingRepairWatcher.AlertBlacklist[blacklistKeyPrefix .. tostring(nPlotMinutesRemaining)] = true
					end
				end
			end
		end
	end
end

function HousingRepairWatcher:CreateExpirationAlert(tPlayerSaveData, tPlotSaveData, strKey)
	-- create new alert and then set its data
	local wAlertItem = Apollo.LoadForm(self.xmlAlert, "frmHousingRepairWatcherAlertItem", self.wndAlert, self)
	local CharacterLabel = wAlertItem:FindChild("frmCharacterTitle")
	local PlotLabel = wAlertItem:FindChild("frmPlotName")
	local ExpirationLabel = wAlertItem:FindChild("frmPlotExpiration")
	
	CharacterLabel:SetText(tPlayerSaveData.strPlayerRealm .. "-" .. tPlayerSaveData.strPlayerName)
	PlotLabel:SetText(tPlotSaveData.strPlotName)
	ExpirationLabel:SetText(HousingRepairWatcher:GetProgressText(tPlotSaveData.fPlotRemainingHours))


	wAlertItem:SetData(strKey)
	self.wndAlert:ArrangeChildrenVert()

	HousingRepairWatcher.CurrentAlertCount = HousingRepairWatcher.CurrentAlertCount + 1

	if self.wndAlert:IsVisible() == false then
		self.wndAlert:Show(true)
	end
end

-----------------------------------------------------------------------------------------------
-- HousingRepairWatcherForm Functions
-----------------------------------------------------------------------------------------------

-- Returns the current character's main data in the save format
function HousingRepairWatcher:CreateOrUpdateCharacterSaveData()

	local player = GameLib.GetPlayerUnit()
	
	-- Grab the current player's data if it already exists
	local tPlayerSaveData = HousingRepairWatcher.AccountCharacters[GameLib.GetRealmName() .. "-" .. player:GetTitleOrName()]
	
	-- If it doesn't already exist, create it
	if tPlayerSaveData == nil then
		tPlayerSaveData = {}
	end
	
	-- Update the player's data
	tPlayerSaveData.nPlayerId = player:GetId()
	tPlayerSaveData.strPlayerName = player:GetTitleOrName()
	tPlayerSaveData.strPlayerRealm = GameLib.GetRealmName()

	if tPlayerSaveData.tPlots == nil then
		tPlayerSaveData.tPlots = {}
	end
	
	return tPlayerSaveData
end

-- Returns the specified plot data in save format
function HousingRepairWatcher:CreateOrUpdatePlotSaveData(udPlotData)
	
	--self:PrintPlotObjectData(udPlotData)
	
	local plotData = {}

	plotData.strPlotName = udPlotData.strName
	plotData.nPlotIndex = udPlotData.nPlotPropertyIndex
	plotData.timestamp = GameLib.GetServerTime()
	plotData.fPlotRemainingHours = udPlotData.fRemainingHours
	plotData.eUpkeepType = udPlotData.eUpkeepType
	plotData.bNeedsRepair = udPlotData.bNeedsRepair	
	plotData.nPlotChargesRemaining = udPlotData.nChargesRemaining

	--self:PrintDebugMessage("End Create or Update Plot")
	return plotData
end

-- Updates the plot display based on the current information
function HousingRepairWatcher:UpdatePlot(nPlotIndex, udPlotData)
	local plotButton = HousingRepairWatcher.wndMain:FindChild(HousingRepairWatcher.PlotControlPrefix .. nPlotIndex)
	local plotNameControl = HousingRepairWatcher.wndMain:FindChild(HousingRepairWatcher.PlotNameControlPrefix .. nPlotIndex)
	local plotRepairBar = HousingRepairWatcher.wndMain:FindChild(HousingRepairWatcher.PlotRepairBarPrefix .. nPlotIndex)
	local plotNeedsRepairSprite = HousingRepairWatcher.wndMain:FindChild(HousingRepairWatcher.PlotOccupiedSpritePrefix .. nPlotIndex)
	local plotTimeRemainingControl = HousingRepairWatcher.wndMain:FindChild(HousingRepairWatcher.PlotTimeRemainingControlPrefix .. nPlotIndex)

	if udPlotData == nil then
		plotNameControl:SetText("No Plot")
		self:UpdateProgressBar(plotRepairBar, 10, 10, "")
	else

		--self:PrintPlotSaveData(udPlotData)
		
		-- Calculate how much time is actually left since the snapshot
		local nPlotHoursRemaining = self:CalculateHoursRemaining(udPlotData.timestamp, udPlotData.fPlotRemainingHours)
		local nChargesRemaining = udPlotData.nPlotChargesRemaining
		local bPlotNeedsRepair = udPlotData.bNeedsRepair
		local nFullyRepairedHours = (7 * 24)
		local strPlotName = udPlotData.strPlotName


		if string.len(strPlotName) > 25 then
			strPlotName = string.sub(udPlotData.strPlotName, 1, 25) .. "..."
		end

		-- Show the plot control
		plotButton:Show(true)

		-- Set the plot's name

		plotNameControl:SetText(strPlotName)
		plotTimeRemainingControl:SetText(HousingRepairWatcher:GetProgressText(nPlotHoursRemaining))

		-- Show repair icon if plot is currently in a state of repair
		if nPlotHoursRemaining == 0 then
			plotNeedsRepairSprite:Show(true, true)
		else
			plotNeedsRepairSprite:Show(false)
		end	
		
		--self:PrintDebugMessage("Calculated Hours Remaining: " .. tostring(nPlotHoursRemaining))

		if udPlotData.eUpkeepType == HousingLib.HousingUpkeepType.Timed then
			-- Set the progress bar based on how much time is left
			self:UpdateProgressBar(plotRepairBar, nFullyRepairedHours, nPlotHoursRemaining, HousingRepairWatcher:GetProgressText(nPlotHoursRemaining))
		elseif udPlotData.eUpkeepType == HousingLib.HousingUpkeepType.Charged then
			-- If it's a charge, just display how many charges are left. Not sure how to
			-- get the total charge yet....
			self:UpdateProgressBar(plotRepairBar, 10, 10, tostring(nChargesRemaining) .. " Charges Remaining")
		elseif udPlotData.eUpkeepType == HousingLib.HousingUpkeepType.Permanent then
			-- If it's a permanent plot, no need to repair
			self:UpdateProgressBar(plotRepairBar, 10, 9.99, "Permanent Plot: No Repairs Needed")
			plotTimeRemainingControl:SetText("N/A")
		elseif udPlotData.eUpkeepType == HousingLib.HousingUpkeepType.Decay then
			-- Decay plot, can't repair
			self:UpdateProgressBar(plotRepairBar, 10, 10, "Decay Plot: Can't Repair")
		elseif udPlotData.eUpkeepType == HousingLib.HousingUpkeepType.TimedCharged then
			-- TimedCharged plot...show time and charges I guess?
			self:UpdateProgressBar(plotRepairBar, nFullyRepairedHours, nPlotHoursRemaining, HousingRepairWatcher:GetProgressText(nPlotHoursRemaining) .. " / " .. tostring(nChargesRemaining) .. " Charges Remaining")
		else
			self:UpdateProgressBar(plotRepairBar, 10, 10, "Structure Plot: Can't Repair")
			plotTimeRemainingControl:SetText("N/A")
		end
	end

	--self:PrintDebugMessage("End Plot Update")
end

-----------------------------------------------------------------------------------------------
-- UI Events
-----------------------------------------------------------------------------------------------
-- when the OK button is clicked
function HousingRepairWatcher:OnOK()
	HousingRepairWatcher.wndMain:Close() -- hide the window
end

-- when the Cancel button is clicked
function HousingRepairWatcher:OnCancel()
	HousingRepairWatcher.wndMain:Close() -- hide the window
end

-- when the user clicks the toggle player list button
function HousingRepairWatcher:btnTogglePlayerListClicked( wndHandler, wndControl, eMouseButton )
	-- Clear the existing player dropdown
	local wPlayerList = HousingRepairWatcher.wndMain:FindChild("wPlayerListFrame")
	wPlayerList:DestroyChildren()

	-- Toggle list based on previous visibility
	if wPlayerList:IsVisible() == true then
		wPlayerList:Show(false)
		return
	-- If the list isn't visible, populate and show it
	else
		wPlayerList:Show(true)
		-- Populate the Player list 
		for key, tPlayerSave in pairs(HousingRepairWatcher.AccountCharacters) do
			local wPlayerItem = Apollo.LoadForm(self.xmlDoc, "frmPlayerListItem", wPlayerList, self)
			-- Set the player name
			wPlayerItem:FindChild("btnPlayerListItem"):SetText(key)
			wPlayerItem:SetData(tPlayerSave)
		end

		wPlayerList:ArrangeChildrenVert()
	end
end

-- when the user clicks on a player in the player list
function HousingRepairWatcher:frmPlayerListItemClicked( wndHandler, wndControl, eMouseButton )
	HousingRepairWatcher.CurrentCharacterKey = wndControl:GetText()
	--self:PrintDebugMessage("Item Clicked: " .. wndControl:GetText())
	-- Close the player list
	local wPlayerList = HousingRepairWatcher.wndMain:FindChild("wPlayerListFrame")
	wPlayerList:DestroyChildren()
	wPlayerList:Show(false)	
end

-- when the user moves the window
function HousingRepairWatcher:frmRepairWatcherFormWindowMove( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
	--[[
	-- get the new position
	local nLeft, nTop, nRight, nBottom = HousingRepairWatcher.wndMain:GetAnchorOffsets()
	HousingRepairWatcher.Settings["nWindowTop"] = nTop
	HousingRepairWatcher.Settings["nWindowRight"] = nRight
	HousingRepairWatcher.Settings["nWindowRight"] = nBottom
	HousingRepairWatcher.Settings["nWindowLeft"] = nLeft
	--]]
end

-----------------------------------------------------------------------------------------------
-- Helper Functions
-----------------------------------------------------------------------------------------------

-- Populates the specified progress bar's properties
function HousingRepairWatcher:UpdateProgressBar(plotRepairBar, nMax, nProgress, strText)
	--plotRepairBar:SetText(strText)
	plotRepairBar:SetText("")
	plotRepairBar:SetMax(nMax)
	plotRepairBar:SetProgress(nProgress)
end

-- Returns the text to display on the progress bar based on the number of hours remaining
function HousingRepairWatcher:GetProgressText(nHoursRemaining)
	-- Show hours if the time left is > 1 hour
	if nHoursRemaining >= 1 then
		return tostring(HousingRepairWatcher:RoundNumber(nHoursRemaining, 2)) .. " Hrs"
	-- Show minutes if there is under an hour left but more than a minute
	elseif nHoursRemaining < 1 and nHoursRemaining * 60 > 1 then
		return tostring(HousingRepairWatcher:RoundNumber(nHoursRemaining * 60, 2)) .. " Min"
	-- Show seconds if there is under a minute left but plot isn't expired
	elseif nHoursRemaining < 1 and nHoursRemaining * 60 < 1 and nHoursRemaining > 0 then
		return tostring(HousingRepairWatcher:RoundNumber(nHoursRemaining * 60 * 60, 2)) .. " Sec"
	else
		return "Repair!"
	end
end

-- Prints the specified debug message if the option to show debug
-- messages is enabled
function HousingRepairWatcher:PrintDebugMessage(strMessage)
	if HousingRepairWatcher.Settings["bDisplayDebugMessages"] ~= nil and HousingRepairWatcher.Settings["bDisplayDebugMessages"].ConfigValue == true then
		Print(strMessage)
	end
end

-- Prints out the data for the specified player save
function HousingRepairWatcher:PrintPlayerSaveData(tPlayerSaveData)
	if tPlayerSaveData == nil then
		self:PrintDebugMessage("[info]Player Save Data nil, Cannot Print")
	else
		self:PrintDebugMessage("[info] Player Data:")
		self:PrintDebugMessage("[info]          ID: " .. tostring(tPlayerSaveData.nPlayerId))
		self:PrintDebugMessage("[info]        Name: " .. tPlayerSaveData.strPlayerName)
		self:PrintDebugMessage("[info]       Realm: " .. tPlayerSaveData.strPlayerRealm)
	end
end

-- Prints out the data for the specified plot save
function HousingRepairWatcher:PrintPlotSaveData(tPlotSaveData)
	if tPlotSaveData == nil then
		self:PrintDebugMessage("[info]Plot Save Data nil, Cannot Print")
	else
		self:PrintDebugMessage("Plot Name: " .. tPlotSaveData.strPlotName)
		self:PrintDebugMessage("       Plot Index: " .. tostring(tPlotSaveData.nPlotIndex))
		self:PrintDebugMessage("        TimeStamp: " .. tostring(tPlotSaveData.timestamp))
		self:PrintDebugMessage("  Remaining Hours: " .. tostring(tPlotSaveData.fPlotRemainingHours))
		self:PrintDebugMessage("      Plot Upkeep: " .. tostring(tPlotSaveData.eUpkeepType))
		self:PrintDebugMessage("      Need Repair: " .. tostring(tPlotSaveData.bNeedsRepair))
		self:PrintDebugMessage(" Charge Remaining: " .. tostring(tPlotSaveData.nPlotChargesRemaining))
	end
end

-- Prints out a plot table returned by the API GetPlot() function
function HousingRepairWatcher:PrintPlotObjectData(tPlotObject)
	if tPlotObject == nil then
		self:PrintDebugMessage("[info]Plot Data nil, Cannot Print")
	else
		self:PrintDebugMessage("Plot Name: " .. tPlotObject.strName)
		self:PrintDebugMessage("  Plot Index: " .. tostring(tPlotObject.nPlotPropertyIndex))
		self:PrintDebugMessage("  Remaining Hours: " .. tostring(tPlotObject.fRemainingHours))
		self:PrintDebugMessage("  Plot Upkeep: " .. tostring(tPlotObject.eUpkeepType))
		self:PrintDebugMessage("  Need Repair: " .. tostring(tPlotObject.bNeedsRepair))
		self:PrintDebugMessage("  Charge Remaining: " .. tostring(tPlotObject.nChargesRemaining))
	end
end

function HousingRepairWatcher:PrintConfiguration(tConfiguration)
	if tConfiguration == nil then
		self:PrintDebugMessage("[info]Configuration Data nil, Cannot Print")
	else
		for key, setting in pairs(HousingRepairWatcher.Configuration) do
			Print(tostring(key) .. ":")
			Print("  Display: " .. setting.ConfigDisplay)
			Print("    Value: " .. tostring(setting.ConfigValue))
		end
	end
end

function HousingRepairWatcher:PrintTable(tTable, nIndent)
	if not nIndent then nIndent = 0 end
	for k, v in pairs(tTable) do
		formatting = string.rep("  ", nIndent) .. k .. ": "
		if type(v) == "table" then
			Print(formatting)
			HousingRepairWatcher:PrintTable(v, nIndent+1)
		elseif type(v) == 'boolean' then
			Print(formatting .. tostring(v))      
		else
			Print(formatting .. v)
		end
	end
end

-- Rounds the specified number to the specified number of decimal places
-- using round-up logic (round up when .5 or higher)
function HousingRepairWatcher:RoundNumber(nValue, nDecimalPlaces)
	local mult = 10^(nDecimalPlaces or 0)
	return math.floor(nValue * mult + 0.5) / mult
end

-- Calculates the hours remaining given the capture time, hours remaining at that time, and the current time
function HousingRepairWatcher:CalculateHoursRemaining(tsCaptureTime, fHoursRemaining)
	--self:PrintDebugMessage("Begin Calculating Hours")

	local startingTime = self:GetEpochTime(tsCaptureTime)
	local currentTime = self:GetEpochTime(GameLib.GetServerTime())

	-- diff is in seconds so multiply by sec/min and min/hour
	local diffInHours = ((currentTime - startingTime) / 60) / 60

	return fHoursRemaining - diffInHours
end

-- Convert timestamp to epoch time (seconds since epoch)
function HousingRepairWatcher:GetEpochTime(tsTime)
	-- os.time *usually* returns seconds since epoch. We're going to assume it's doing so here as well
	return os.time({year = tsTime.nYear, month = tsTime.nMonth, day = tsTime.nDay, hour = tsTime.nHour, min = tsTime.nMinute, sec = tsTime.nSecond})
end

-- Returns a string in format MM-dd-yyyy HH:MM:SI
function HousingRepairWatcher:ConvertTimestampToString(tsTime)
	local strFormattedTimestamp = tostring(tsTime.nMonth) .. "-" .. tostring(tsTime.nDay) .. "-" .. tostring(tsTime.nYear) .. " " .. tostring(tsTime.nHour) .. ":" .. tostring(tsTime.nMinute) .. ":" .. tostring(tsTime.nSecond)
	return strFormattedTimestamp
end

-----------------------------------------------------------------------------------------------
-- HousingRepairWatcher Instance
-----------------------------------------------------------------------------------------------
local HousingRepairWatcherInst = HousingRepairWatcher:new()
HousingRepairWatcherInst:Init()
